/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	7DRL Development
 *
 * NAME:        gfxengine.h ( Live Once Library, C++ )
 *
 * COMMENTS:
 *	Provides the common interface to the screen display functions
 *	and input handling functions.
 */

#ifndef __gfxengine__
#define __gfxengine__

#define SCR_WIDTH	80
#define SCR_HEIGHT	60
#define MAX_LAYER	6
#define SCR_RAD		8
#define LOWER_LAYERS	5
#define MAX_LEVELS	5

#include "glbdef.h"

// Key defines.
enum GFX_Keydefine
{
    GFX_KEYSTART = 256,
    GFX_KEYUP,
    GFX_KEYLEFT,
    GFX_KEYRIGHT,
    GFX_KEYDOWN,
    GFX_KEYF1,
    GFX_KEYF2,
    GFX_KEYF3,
    GFX_KEYF4,
    GFX_KEYF5,
    GFX_KEYF6,
    GFX_KEYF7,
    GFX_KEYF8,
    GFX_KEYF9,
    GFX_KEYF10,
    GFX_KEYF11,
    GFX_KEYF12,
    GFX_KEYSF1,
    GFX_KEYSF2,
    GFX_KEYSF3,
    GFX_KEYSF4,
    GFX_KEYSF5,
    GFX_KEYSF6,
    GFX_KEYSF7,
    GFX_KEYSF8,
    GFX_KEYSF9,
    GFX_KEYSF10,
    GFX_KEYSF11,
    GFX_KEYSF12,
    GFX_KEYLAST
};

enum GFX_FadeMode
{
    GFX_FADE_NONE,
    GFX_FADE_BLACK,
    GFX_FADE_WHITE,
    NUM_GFX_FADES
};

enum JUST_Horizontal
{
    JUST_LEFT,
    JUST_CENTER,
    JUST_RIGHT
};
enum JUST_Vertical
{
    JUST_TOP,
    JUST_BOTTOM
};

struct gfx_bmpoverlay
{
    u8 		*rgbdata;
    bool	transparent;
    int		scale;
    int		layer;
    int		x, y;
    int		w, h;
};

typedef	unsigned char u8;

void gfx_init();
void gfx_shutdown();

void gfx_delay(int frames);

void gfx_setfade(GFX_FadeMode mode, u8 amount);
void gfx_fadeto(GFX_FadeMode mode, bool increase);
void gfx_shake(int duration);
void gfx_setscrolloffset(int dx, int dy);
void gfx_getscrolloffset(int &dx, int &dy);

void gfx_printchar(int x, int y, u8 c, ATTR_NAMES attr);
void gfx_printstring(int x, int y, const char *s, ATTR_NAMES attr);

void gfx_justifystring(int x, int y, JUST_Horizontal jh, JUST_Vertical jv, 
			const char *s, ATTR_NAMES attr);

void gfx_clearbox(int x, int y, int w, int h, u8 c, ATTR_NAMES attr);

enum
{
    GFX_OVERLAY_BACK,		// Behind all
    GFX_OVERLAY_MID,		// On top of map, under text and scrolls
    GFX_OVERLAY_TOP		// On top of all, not faded.
};

// Creates an overlay bitmap and returns a handle to it.
// Makes a copy of the rgb data!
int gfx_overlay_create(const u8 *rgb, int w, int h,
			int x, int y, int scale, 
			int layer,
			bool transparent);
void gfx_overlay_center(int handle);
void gfx_overlay_destroy(int handle);

void gfx_getchar(int x, int y, u8 &c, ATTR_NAMES &attr);

void gfx_settile(int layer, int x, int y, TILE_NAMES tile, int offx = 0, int offy = 0);
TILE_NAMES gfx_gettile(int layer, int x, int y, int &offx, int &offy);
TILE_NAMES gfx_gettile(int layer, int x, int y);

void gfx_settile_light(int x, int y, int light);
int gfx_gettile_light(int x, int y);

void gfx_settile_level(int x, int y, int level);
int gfx_gettile_level(int x, int y);

// Nothing is drawn until this is called.
void gfx_update();

// True if a key is waiting on the input queue.
bool gfx_isKeyWaiting();

// This empties the keyboard buffer without blocking.
void gfx_clearKeyBuffer();

// This attempts to end any current key repeating.
void gfx_breakKeyRepeat();

// Returns 0 for no key.  ASCII for ascii keys.
// Other stuff gets interesting.
// If block is true, will wait for a key press.
int gfx_getKey(bool block = true);

// Inputs a string, reporting it to the given screen location,
// buf is always null terminated and is of length maxlen.  Up to maxlen-1
// characters will be read.
void gfx_getString(int x, int y, ATTR_NAMES attr, char *buf, int maxlen);

// Allocates a buffer for the bitmap.  Converts it to 24 bit
// rgb format.  w & h are set to the size.
u8 *gfx_loadbitmap(const char *fname, int &w, int &h);

// Loads the bitmap but also looks up override files.
u8 *gfx_loadbitmap_override(const char *fname, int level, int &w, int &h);

#endif

