#include "item.h"
#include "map.h"

ITEM::ITEM(ITEM_NAMES _defn)
{
    myDefinition = _defn;
    myPos = POS::invalidPos();
    myFacing = DIR::north();
    myBaseMob = MOB_NONE;
    myHP = defn().maxhp;
}

ITEM::~ITEM()
{
    if (myPos.map())
	myPos.map()->deregisterItem(this);
}

TILE_NAMES
ITEM::tile() const
{
    int		x, y;
    return tile(x, y);
}

TILE_NAMES
ITEM::tile(int &toffx, int &toffy) const
{
    TILE_NAMES		itemtile = (TILE_NAMES) defn().tile;

    if (getDefinition() == ITEM_CORPSE)
    {
	int		facing = getFacing().angle();
	MOB_NAMES	corpsemob = getBaseMob();

	itemtile = (TILE_NAMES) glb_mobdefs[corpsemob].tile;
	toffx = facing;
	toffy = 4;
	return itemtile;
    }

    toffx = 0;
    toffy = 0;

    if (glb_tiledefs[itemtile].randset_x || glb_tiledefs[itemtile].randset_y)
    {
	int		rx, ry;
	rx = rand_wanginthash(rand_wanginthash(pos().x()) + pos().y());
	rx += pos().y();
	rx = rand_wanginthash(rx);
	ry = rand_wanginthash(rx);
	rx = abs(rx) % glb_tiledefs[itemtile].randset_x;
	ry = abs(ry) % glb_tiledefs[itemtile].randset_y;
	if (glb_tiledefs[itemtile].doublehigh)
	    ry *= 2;

	toffx = rx;
	toffy = ry;
    }

    if (glb_tiledefs[itemtile].durability)
    {
	// Make room for multiple randsets.
	toffx *= 3;
	if (myHP == defn().maxhp)
	{
	    // Mint.
	}
	else if (myHP == 1)
	{
	    // Almost dead
	    toffx += 2;
	}
	else
	{
	    // Dogged
	    toffx += 1;
	}
    }
    return itemtile;
}

void
ITEM::moveTo(POS newpos)
{
    if (myPos.map())
    {
	if (newpos.map() != myPos.map())
	{
	    myPos.map()->deregisterItem(this);
	    if (newpos.map())
	    {
		myPos = newpos;
		newpos.map()->registerItem(this);
	    }
	}
	else
	{
	    POS		oldpos = myPos;
	    myPos = newpos;
	    myPos.map()->updateItemCache(this, oldpos, newpos);
	}
    }
    else if (newpos.map())
    {
	myPos = newpos;
	newpos.map()->registerItem(this);
    }
    myPos = newpos;
}

bool
ITEM::heartbeat(int phase)
{
    return true;
}

bool
ITEM::heal(int h)
{
    myHP += h;
    if (myHP > defn().maxhp)
	myHP = defn().maxhp;
    return true;
}

bool
ITEM::damage(int damage)
{
    myHP -= damage;
    if (myHP <= 0)
    {
	myHP = 0;
	return false;
    }
    return true;
}
