#include "mob.h"
#include "map.h"
#include "dist.h"
#include "atlas.h"
#include "text.h"
#include "gfxengine.h"
#include "runes.h"

extern void readBook(const char *text);
extern void readBook(BUF text);
extern void fadeToBlack();
extern void fadeFromBlack();

extern void fadeToWhite();
extern void redrawWorld();

extern bool glbVictory;

MOB::MOB(MOB_NAMES _defn)
{
    myDefinition = _defn;
    myPos = POS::invalidPos();
    myDiskSourceMap = 0;
    myLastPhase = 0;
    myFacing = DIR::north();
    myLastWalkDir = DIR::north();
    myIsAttacking = false;
    myHP = defn().maxhp;
    mySelection = -1;

#if 0
    if (_defn == MOB_AVATAR)
    {
	gainItem(new ITEM(ITEM_BOOK_SHADOW));
    }
#endif

    for (const char *i = MOB::defn().intrinsic; *i; i++)
    {
	intrinsic().set((INTRINSIC_NAMES) *i);
    }
}

MOB::~MOB()
{
    if (myPos.map())
	myPos.map()->deregisterMob(this);
}

void
MOB::moveTo(POS newpos)
{
    if (myPos.map())
    {
	if (newpos.map() != myPos.map())
	{
	    myPos.map()->deregisterMob(this);
	    if (newpos.map())
	    {
		myPos = newpos;
		newpos.map()->registerMob(this);
	    }
	}
	else
	{
	    myPos.map()->updateMobCache(this, myPos, newpos);
	}
    }
    else if (newpos.map())
    {
	myPos = newpos;
	newpos.map()->registerMob(this);
    }
    myPos = newpos;
}

bool
MOB::actionPickUp()
{
    if (!pos().item())
    {
	return false;
    }

    if (!pos().item()->defn().inventory)
    {
	actionEmote("*too heavy*");
	return false;
    }

    //actionEmote("*takes*");
    gainItem(pos().item());
    return true;
}

bool
MOB::actionInvokeSelected()
{
    if (!numItems())
    {
	actionEmote("*waves hands*");
	return false;
    }
    ITEM		*selected = inventory(mySelection);

    if (selected->getDefinition() == ITEM_DAYSTAR)
    {
	// Place it in the square in front of us.
	POS		placepos = pos().delta(getFacing());

	if (placepos.item())
	{
	    actionEmote("*no room*");
	    return false;
	}

	if (pos().item() && pos().item()->getDefinition() == ITEM_ALTAR)
	{
	    if (getFacing() == DIR::north())
	    {
		// actionEmote("*release*");
		loseItem(selected);
		selected->moveTo(placepos);

		fadeToWhite();

		glbAtlas()->addDaystar();

		return true;
	    }
	}

	if (!placepos.passable())
	{
	    actionEmote("*no room*");
	    return false;
	}

	loseItem(selected);
	selected->moveTo(placepos);
	placepos.removeDarkness();
	placepos.north().removeDarkness();
	placepos.west().removeDarkness();
	placepos.east().removeDarkness();
	placepos.south().removeDarkness();
	return true;
    }

    if (selected->defn().spell != SPELL_NONE)
    {
	actionCast((SPELL_NAMES) selected->defn().spell);

	if (!selected->damage(1))
	{
	    loseItem(selected);
	    delete selected;
	}

	return true;
    }
    else
    {
	// Not a book?  Confused.
	actionEmote("*brandishes*");
	return false;
    }

    return false;
}

bool
MOB::actionBump(DIR dir)
{
    POS  p = pos().delta(dir);

    // Check if we are facing the right direction.
    if (dir != getFacing())
    {
	return actionTurnTo(dir);
    }

    if (p.mob())
    {
	return actionAttack(dir);
    }

    if (isAvatar())
    {
	ITEM		*item = p.item();
	if (item)
	{
	    switch (item->getDefinition())
	    {
		case ITEM_WILDGROWTH:
		case ITEM_WILDGROWTH_DAMAGED:
		{
		    // Hack it!
		    return actionAttack(dir);
		}
		case ITEM_PORTAL_TOP:
		{
		    // Win the game!
		    glbVictory = true;

		    return true;
		}
		case ITEM_BOULDER:
		{
		    // Attempt to push!
		    POS		nextpos = p.delta(dir);
		    if (nextpos.item() || !nextpos.passable())
		    {
			// Swap
			//actionEmote("*swap*");
			item->moveTo(pos());
			return actionWalk(dir);
		    }
		    else
		    {
			//actionEmote("*push*");
			item->moveTo(nextpos);
			return actionWalk(dir);
		    }
		}
		default:
		    break;
	    }
	}
    }
    if (!canMoveIn(dir))
    {
	// Check for block actions...
	if (isAvatar())
	{
	    switch (p.tile())
	    {
		case SQUARE_DOOR:
		    actionEmote("*opens*");
		    p.setTile(SQUARE_DOOR_OPEN);
		    return true;
		case SQUARE_WALL_LEVER_DOWN:
		case SQUARE_WALL_LEVER_UP:
		    actionEmote("*flips*");
		    p.toggleLinkedSquares();
		    return true;
		default:
		    break;
	    }


	}
	// No slide, blocked
	return false;
    }

    return actionWalk(dir);
}

void
MOB::damage(const char *attack, int hurt)
{
    myHP -= hurt;
    if (myHP <= 0)
    {
	myHP = 0;
	die(attack);
    }
}

void
MOB::die(const char *death)
{
    if (isAvatar())
    {
	// We do nothing, and let us actually be killed
	// by the main loop wehre it is safer...
	// In mean time, hp == 0 will give us death animation
	myHP = 0;
    }
    else
    {
	// Create corpse object.
	ITEM		*corpse = new ITEM(ITEM_CORPSE);
	corpse->setFacing(getFacing());
	corpse->setBaseMob(getDefinition());
	corpse->moveTo(pos());
	moveTo(POS::invalidPos());
	delete this;
    }
}

bool
MOB::actionAttack(DIR dir)
{
    POS		newpos = pos().delta(dir);
    MOB		*mob = newpos.mob();
    ITEM	*item = newpos.item();

    if (dir != getFacing())
	return actionTurnTo(dir);

    if (!mob)
    {
	if (item && isAvatar())
	{
	    if (item->getDefinition() == ITEM_WILDGROWTH)
	    {
		// actionEmote("*stab*");

		setAttacking(true);
		redrawWorld();
		gfx_delay(10);
		setAttacking(false);
		redrawWorld();
		item->setDefinition(ITEM_WILDGROWTH_DAMAGED);
		return true;
	    }
	    else if (item->getDefinition() == ITEM_WILDGROWTH_DAMAGED)
	    {
		// actionEmote("*stab*");

		setAttacking(true);
		redrawWorld();
		gfx_delay(10);
		setAttacking(false);
		redrawWorld();

		item->moveTo(POS::invalidPos());
		delete item;
		return true;
	    }
	}

	return false;
    }

    if (isAvatar())
    {
	// actionEmote("*stab*");

	setAttacking(true);
	redrawWorld();
	gfx_delay(10);
	setAttacking(false);
	redrawWorld();

	mob->damage(defn().attack, 1);
	return true;
    }

    {
	BUF		attack;
	setAttacking(true);
	attack.sprintf("*%s*", defn().attack);
	// actionEmote(attack.buffer());
	mob->damage(defn().attack, 1);
    }

    return true;
}

bool
MOB::actionTurnTo(DIR dir)
{
    setFacing(dir);

    // Yep, takes time!
    return true;
}

bool
MOB::actionWalk(DIR dir)
{
    POS		newpos = pos().delta(dir);

    if (!canMoveTo(newpos))
	return false;

    // Check if we are facing the right direction.
    if (dir != getFacing())
    {
	return actionTurnTo(dir);
    }

    myLastWalkDir = dir;

    // Actually move first so we are shown on it.
    moveTo(newpos);

    // See if there is anything special here!
    ITEM		*item = newpos.item();
    if (item)
    {
	switch (item->getDefinition())
	{
	    case ITEM_BED:
		if (isAvatar())
		{
		    actionEmote("*zzz*");
		    fadeToBlack();
		    readBook(text_lookup("item", item->defn().name));
		    fadeFromBlack();
		}
		break;
	}

	// See if we can pick it up.
	if (isAvatar() && item->defn().inventory)
	{
	    return actionPickUp();
	}
    }

    switch (newpos.tile())
    {
	case SQUARE_ENTRANCEDOOR:
	{
	    actionEmote("*leaves*");
	    MAP		*map = glbAtlas()->findMap("world");
	    actionEmote("*blocked*");
	    break;
	}
	case SQUARE_FLOOR_TELE:
	case SQUARE_GROUND_TELE:
	{
	    actionEmote("*teleports*");
	    POSLIST		destlist;
	    newpos.findLinkedSquares(destlist);
	    for (int i = 0; i < destlist.entries(); i++)
	    {
		if (destlist(i).x() != newpos.x() ||
		    destlist(i).y() != newpos.y())
		{
		    newpos = destlist(i);
		    break;
		}
	    }
	    break;
	}
	case SQUARE_FLOOR_SPINNER:
	case SQUARE_GROUND_SPINNER:
	case SQUARE_FOREST_SPINNER:
	{
	    actionEmote("*dizzy*");
	    setFacing(DIR::rand());
	    break;
	}
	case SQUARE_FLOOR_HOLE_TRAP:
	case SQUARE_GROUND_HOLE_TRAP:
	case SQUARE_FLOOR_HOLE:
	case SQUARE_GROUND_HOLE:
	{
	    POS		p = newpos.down();
	    if (p.valid())
	    {
		actionEmote("*falls*");
		newpos = p;
		// Check to see if we have feather fall
		if (intrinsic().get(INTRINSIC_FEATHERFALL))
		{
		    intrinsic().clear(INTRINSIC_FEATHERFALL);
		}
		else
		{
		    die("fall");
		    return true;
		}
	    }
	    else
		actionEmote("*hole is blocked*");
	    break;
	}
	case SQUARE_WATER:
	{
	    if (!intrinsic().get(INTRINSIC_WATERWALK))
	    {
		actionEmote("*drowns*");
		die("drown");
		return true;
	    }
	    intrinsic().clear(INTRINSIC_WATERWALK);
	    break;
	}
	case SQUARE_LAVA:
	{
	    if (!intrinsic().get(INTRINSIC_RESISTFIRE))
	    {
		actionEmote("*burns*");
		die("burn");
		return true;
	    }
	    intrinsic().clear(INTRINSIC_RESISTFIRE);
	    if (!intrinsic().get(INTRINSIC_WATERWALK))
	    {
		actionEmote("*drowns*");
		die("drown");
		return true;
	    }
	    intrinsic().clear(INTRINSIC_WATERWALK);
	    break;
	}
    }
    // Move again in case of a double move
    moveTo(newpos);
    return true;
}

bool
MOB::actionYell(const char *msg)
{
    if (defn().speaker != MOB_NONE &&
	defn().speaker != getDefinition())
    {
	MOB	*speaker = pos().map()->findMobByType((MOB_NAMES)defn().speaker);
	if (speaker)
	    return speaker->actionYell(msg);
    }

    BUF		yell;
    yell.strcpy(msg);
    yell.stripWS();

    if (!yell.isstring())
    {
	// Empty yell
	return false;
    }
    myMessage.strcpy(yell);

    return true;
}

void
MOB::gainItem(ITEM *item)
{
    item->moveTo(POS::invalidPos());
    myInventory.append(item);
    if (mySelection < 0)
	mySelection = 0;
}

void
MOB::loseItem(ITEM *item)
{
    assert(myInventory.find(item) >= 0);
    myInventory.removePtr(item);
    if (mySelection >= myInventory.entries())
    {
	mySelection = myInventory.entries()-1;
    }
}

bool
MOB::actionShiftPrimary()
{
    if (!myInventory.entries())
	return false;

    mySelection++;
    if (mySelection >= myInventory.entries())
	mySelection = 0;
    return false;
}

bool
MOB::actionCast(SPELL_NAMES spell)
{
    switch (spell)
    {
	case NUM_SPELLS:
	case SPELL_NONE:
	    break;

	case SPELL_SHADOW:
	{
	    actionYell(glb_spelldefs[spell].words);
	    WAVEFRONT		front(pos(), getFacing());

	    for (int i = 0; i < 5; i ++)
	    {
		for (int j = 0; j < front.entries(); j++)
		{
		    front(j).setDarkness();
		}
		front.advance();
	    }
	    break;
	}
	case SPELL_WILDGROWTH:
	{
	    actionYell(glb_spelldefs[spell].words);

	    POS		basepos = pos().delta(getFacing());
	    DIR		perp;
	    perp = getFacing().clock();
	    for (int i = 0; i < 3; i++)
	    {
		if (!basepos.passable())
		    break;
		if (!basepos.mob() && !basepos.item())
		{
		    ITEM	*i = new ITEM(ITEM_WILDGROWTH);
		    i->moveTo(basepos);
		}
		basepos = basepos.delta(perp);
	    }

	    perp = getFacing().counterclock();
	    basepos = pos().delta(getFacing());
	    for (int i = 0; i < 3; i++)
	    {
		if (!basepos.passable())
		    break;
		if (!basepos.mob() && !basepos.item())
		{
		    ITEM	*i = new ITEM(ITEM_WILDGROWTH);
		    i->moveTo(basepos);
		}
		basepos = basepos.delta(perp);
	    }
	    break;
	}
	case SPELL_VOLCANO:
	{
	    actionYell(glb_spelldefs[spell].words);
	    POS		basepos = pos().delta(getFacing());
	    basepos = basepos.delta(getFacing());
	    basepos.convertToLava();
	    break;
	}
	case SPELL_RECALL:
	{
	    actionYell(glb_spelldefs[spell].words);
	    POS		surface = glbAtlas()->findMap("world")->findRandomPassable();
	    moveTo(surface);
	    break;
	}
	case SPELL_SHADOWSTEP:
	{
	    actionYell(glb_spelldefs[spell].words);

	    POS		dest = pos().delta(getFacing()).delta(getFacing()).delta(getFacing());
	    if (dest.passable() && !dest.mob())
	    {
		moveTo(dest);
		dest.setDarkness();
		if (dest.item() && dest.item()->defn().inventory)
		    actionPickUp();
	    }
	    break;
	}
	case SPELL_RESTORE:
	{
	    actionYell(glb_spelldefs[spell].words);
	    for (int i = 0; i < numItems(); i++)
	    {
		if (inventory(i)->getDefinition() != ITEM_BOOK_RESTORE)
		{
		    inventory(i)->heal(inventory(i)->defn().maxhp);
		}
	    }

	    break;
	}
	case SPELL_CAVEIN:
	{
	    actionYell(glb_spelldefs[spell].words);
	    POS		basepos = pos().delta(getFacing());

	    gfx_shake(15);

	    for (int i = 0; i < 3; i++)
	    {
		if (!basepos.passable())
		    break;
		if (basepos.mob())
		{
		    basepos.mob()->die("crush");
		}
		ITEM	*item = new ITEM(ITEM_BOULDER);
		item->moveTo(basepos);

		basepos = basepos.delta(getFacing());
	    }
	    break;
	}
    }
    return true;
}

bool
MOB::actionEmote(const char *msg)
{
    if (defn().speaker != MOB_NONE &&
	defn().speaker != getDefinition())
    {
	MOB	*speaker = pos().map()->findMobByType((MOB_NAMES)defn().speaker);
	if (speaker)
	    return speaker->actionEmote(msg);
    }
    myEmote.strcpy(msg);

    return true;
}

extern MOB *glbAvatar;

bool
MOB::heartbeat(int phase)
{
    // Avoid double moves caused by crossing map boundaries.
    if (phase == myLastPhase)
	return false;

    myLastPhase = phase;

    // Check for burning
    if (pos().getFlag(FLAG_FIRE) ||
	pos().defn().burning)
    {
	if (intrinsic().get(INTRINSIC_RESISTFIRE))
	    actionEmote("*resists*");
	else
	{
	    actionEmote("*burns*");
	    
	    die("burn");
	    return false;
	}
    }

    intrinsic().heartbeat();

    for (int i = myInventory.entries(); i --> 0; )
    {
	if (myInventory(i))
	    myInventory(i)->heartbeat(phase);
    }

    return true;
}

bool
MOB::doAI()
{
    setAttacking(false);
    if (isAvatar())
	return false;

    if (defn().ai != AI_CHARGE)
	return false;

    bool		avatarvisible = true;
    if (!glbAvatar)
	avatarvisible = false;
    if (avatarvisible && !pos().getFlag(FLAG_FOV))
	avatarvisible = false;
    if (avatarvisible && glbAvatar->intrinsic().get(INTRINSIC_INVIS))
	avatarvisible = false;
    if (avatarvisible && glbAvatar->pos().getFlag(FLAG_DARK))
    {
	// Check actual distance.
	// We can see 1 square around even in shadow.
	if ( MAX(ABS(glbAvatar->pos().x() - pos().x()), ABS(glbAvatar->pos().y() - pos().y())) > 1)
	{
	    avatarvisible = false;
	}
    }

    if (avatarvisible)
    {
	DISTMAP		dist(glbAvatar->pos());

	DIR		d;
	if (dist.grad(pos(), d))
	{
	    // Check if we are beside!
	    if (pos().delta(d).mob() == glbAvatar)
	    {
		return actionAttack(d);
	    }
	    if (actionWalk(d))
		return true;
	}
    }

    return aiWander();
}

bool
MOB::aiWander()
{
    DIR		dir, ndir;
    int		valid = 0;

    dir = getFacing();
    if (canMoveIn(dir))
    {
	if (rand_chance(75))
	    return actionWalk(dir);
    }

    FORALL_DIR(dir)
    {
	if (canMoveIn(dir))
	{
	    if (!rand_choice(++valid))
		ndir = dir;
	}
    }

    if (valid)
    {
	return actionWalk(ndir);
    }

    return false;
}

bool
MOB::canMoveTo(POS newpos) const
{
    if (newpos.mob())
	return false;

    if (!newpos.defn().passable)
	return false;

    // Monsters are smart enough to not suicide.
    if (!isAvatar())
    {
	if (newpos.defn().dangerous)
	    return false;

	if (newpos.getFlag(FLAG_FIRE))
	    return false;

	// They must stay in their square.
	if (newpos.tile() != SQUARE_FLOOR)
	    return false;

	ITEM	*item = newpos.item();

	if (item && item->defn().blockmobs)
	{
	    return false;
	}
    }

    return true;
}
