#include "pos.h"
#include "map.h"
#include "atlas.h"
#include <stdio.h>

SQUARE_NAMES	
POS::tile() const
{
    if (!myMap)
	return SQUARE_NONE;
    return myMap->getTile(myX, myY);
}


int			
POS::level() const 
{ 
    if (!myMap) return 0;
    return map()->level(); 
}

ITEM *	
POS::item() const
{
    if (!myMap)
	return 0;
    return myMap->getItem(myX, myY);
}

ITEM_NAMES
POS::itemname() const
{
    ITEM	*i = item();
    if (!i)
	return ITEM_NONE;
    return i->getDefinition();
}

int
POS::itemlist(ITEMLIST &list) const
{
    list.clear();
    if (!myMap)
	return 0;
    return myMap->getItemList(myX, myY, list);
}

void 		
POS::setTile(SQUARE_NAMES square)
{
    assert(valid());
    if (!valid())
	return;

    myMap->setTile(myX, myY, square);
}

bool
POS::toggleSquare()
{
    if (defn().togglesquare != SQUARE_NONE)
    {
	setTile((SQUARE_NAMES) defn().togglesquare);
	return true;
    }
    return false;
}

void
POS::getLinks(PTRLIST<int> &links) const
{
    links.clear();
    map()->getLinks(myX, myY, links);
}

bool
POS::surface() const
{
    if (!map())
	return false;
    return map()->isSurface();
}

void
POS::findLinkedSquares(POSLIST &list)
{
    list.clear();
    PTRLIST<int>	links;

    map()->getLinks(myX, myY, links);

    for (int i = 0; i < links.entries(); i++)
    {
	POSLIST		plist;
	map()->findLinks(links(i), plist);
	list.append(plist);
    }
}

bool
POS::toggleLinkedSquares()
{
    POSLIST		plist;

    findLinkedSquares(plist);
    bool		didsomething = false;

    for (int j = 0; j < plist.entries(); j++)
    {
	didsomething |= plist(j).toggleSquare();
    }

    return didsomething;
}

bool	
POS::getFlag(FLAG_NAMES flag) const
{
    if (!myMap)
	return SQUARE_NONE;
    return myMap->getFlag(flag, myX, myY);
}

void 		
POS::setFlag(FLAG_NAMES flag, bool val)
{
    assert(valid());
    if (!valid())
	return;

    myMap->setFlag(flag, myX, myY, val);
}

const SQUARE_DEF &
POS::defn() const
{
    return glb_squaredefs[tile()];
}

bool		 
POS::valid() const 
{ 
    return myMap != 0 && myMap->validPos(myX, myY); 
}

MOB *
POS::mob() const
{
    if (!myMap)
	return 0;
    return myMap->getMob(myX, myY);
}

POS
POS::up() const
{
    POS		p = invalidPos();
    if (!valid())
	return p;
    MAP		*goal = map()->atlas()->findMap(map()->defn().up);
    if (!goal)
	return p;
    p = goal->buildPos(x(), y());
    goal->makePosValid(p);
    return p;
}

POS
POS::down() const
{
    POS		p = invalidPos();
    if (!valid())
	return p;
    MAP		*goal = map()->atlas()->findMap(map()->defn().down);
    if (!goal)
	return p;
    p = goal->buildPos(x(), y());
    goal->makePosValid(p);
    return p;
}

POS
POS::delta(DIR dir) const
{
    POS		newpos = *this;

    switch (dir.angle())
    {
	case 0:
	    newpos.myY--;
	    break;
	case 1:
	    newpos.myX--;
	    break;
	case 2:
	    newpos.myY++;
	    break;
	case 3:
	    newpos.myX++;
	    break;
    }

    if (newpos.map())
	newpos.map()->makePosValid(newpos);

    return newpos;
}

void
POS::convertToLava()
{
    if (!valid())
	return;
    // No-op if already lava.
    if (tile() == SQUARE_LAVA)
	return;

    if (defn().lavarisk)
    {
	map()->setTile(x(), y(), SQUARE_LAVA);

	// Kill any mob.
	if (mob())
	{
	    mob()->die("burn");
	}
	ITEMLIST		items;
	itemlist(items);
	for (int i = 0; i < items.entries(); i++)
	{
	    if (items(i)->defn().burnable)
	    {
		items(i)->moveTo(POS::invalidPos());
		delete items(i);
	    }
	}
    }
}

void
POS::setDarkness()
{
    if (tile() == SQUARE_SKY)
	return;
    if (itemname() == ITEM_DAYSTAR)
	return;
    setFlag(FLAG_DARK, true);
}

void
POS::removeDarkness()
{
    setFlag(FLAG_DARK, false);
}

POS
PATH::apply(POS p) const
{
    for (int i = 0; i < myPath.entries(); i++)
    {
	p = p.delta(myPath(i));
    }
    return p;
}

WAVEFRONT::WAVEFRONT(POS center, DIR dir, bool inclusive)
{
    POS		np;
    
    np = center.delta(dir).delta(dir.counterclock());
    if (!np.valid())
	np = center.delta(dir.counterclock()).delta(dir);
    myFront.append(np);

    myFront.append(center.delta(dir));

    np = center.delta(dir).delta(dir.clock());
    if (!np.valid())
	np = center.delta(dir.clock()).delta(dir);
    myFront.append(np);

    myDir = dir;
    myInclusive = inclusive;
    if (myInclusive)
	myFront.append(center.delta(myDir.clock()));
}

void
WAVEFRONT::advance()
{
    POS		p;

    for (int i = 0; i < entries(); i++)
    {
	p = myFront(i);

	myFront.set(i, p.delta(myDir));
    }

    POS np;
    np = p.delta(myDir).delta(myDir.clock());
    if (!np.valid())
	np = p.delta(myDir.clock()).delta(myDir);
    myFront.append(np);
    myFront.insert(0, myFront(0).delta(myDir.counterclock()));
}

CIRCLEFRONT::CIRCLEFRONT(POS center)
{
    myCenter = center;
    DIR dir;
    FORALL_DIR(dir)
    {
	myFronts[dir.angle()] = WAVEFRONT(center, dir);
    }
    myAtCenter = true;
}

POS
CIRCLEFRONT::operator()(int idx) const
{
    if (myAtCenter)
	return myCenter;
    DIR dir;
    FORALL_DIR(dir)
    {
	if (idx < myFronts[dir.angle()].entries())
	    return myFronts[dir.angle()](idx);
	idx -= myFronts[dir.angle()].entries();
    }
    return POS::invalidPos();
}

int
CIRCLEFRONT::entries() const
{
    if (myAtCenter)
    {
	return 1;
    }
    DIR dir;
    int	idx = 0;

    FORALL_DIR(dir)
    {
	idx += myFronts[dir.angle()].entries();
    }
    return idx;
}

void
CIRCLEFRONT::advance()
{
    if (myAtCenter)
    {
	myAtCenter = false;
	return;
    }
    DIR dir;

    FORALL_DIR(dir)
    {
	myFronts[dir.angle()].advance();
    }
}
