#include "runes.h"

#include "rand.h"

RUNES		glbRunes;

static const char *const glbRuneList[26] =
{
    "ad",
    "boi",
    "coa",
    "dex",
    "eim",
    "fla",
    "gad",
    "hap",
    "ick",
    "joka",
    "kana",
    "lat",
    "mef",
    "nal",
    "oup",
    "pila",
    "qut",
    "ret",
    "soph",
    "tink",
    "un",
    "viss",
    "wi",
    "xi",
    "yiy",
    "zizs"
};

BUF
RUNEWORD::buildRune() const
{
    BUF		result;

    for (int i = 0; runes[i]; i++)
    {
	if (i)
	    result.strcat(" ");
	result.strcat(glbRuneList[runes[i]-'a']);
    }

    return result;
}

int
RUNEWORD::usageMask() const
{
    int		mask = 0;

    for (int i = 0; runes[i]; i++)
    {
	mask |= 1 << (runes[i]-'a');
    }

    return mask;
}

static bool
buildRuneList(RUNEWORD *list, int nvalid, int len, int wordsize)
{
    // Pick a new runeword for list[nvalid].

    if (nvalid >= len)
	return true;

    RUNEWORD	word;
    int		symbols[26];

    for (int i = 0; i < 26; i++)
	symbols[i] = i;

    // While this could fail, we are relying on picking a random
    // subset being unlikely to conflict.
    while (1)
    {
	rand_shuffle(symbols, 26);

	for (int i = 0; i < wordsize; i++)
	{
	    word.runes[i] = symbols[i] + 'a';
	}

	// Verify we aren't a match of any other rune.
	int		mask = word.usageMask();

	bool		found = false;
	for (int i = 0; i < nvalid; i++)
	{
	    if (mask == list[i].usageMask())
	    {
		found = true;
		break;
	    }
	}

	if (!found)
	    break;
    }

    list[nvalid] = word;
    return buildRuneList(list, nvalid+1, len, wordsize);
}

RUNES::RUNES()
{
}

RUNES::~RUNES()
{
}

void
RUNES::init()
{
    // We hope we eventually find something valid.
    while (!buildRuneList(mySpell, 0, NUM_SPELLS, 3));

    buildRuneList(&myRiddle, 0, 1, 5);
}

BUF
RUNES::spellRune(SPELL_NAMES spell) const
{
    return mySpell[spell].buildRune();
}

BUF
RUNES::riddleRune() const
{
    return myRiddle.buildRune();
}

SPELL_NAMES
RUNES::findSpell(const char *s) const
{
    BUF		cmp;
    cmp.strcpy(s);
    cmp.stripWS();

    SPELL_NAMES		spell;
    FOREACH_SPELL(spell)
    {
	if (!cmp.strcasecmp(spellRune(spell)))
	    return spell;
    }
    return SPELL_NONE;
}

RUNES *
RUNES::getRunes()
{
    return &glbRunes;
}
